@@ -9,14 +9,155 @@ using Mono.Debugger.Languages;
 
 namespace Mono.Debugger.Frontend
 {
-	public enum LocationType
+	public class ExpressionParser : IExpressionParser
 	{
-		Method,
-		DelegateInvoke,
-		PropertyGetter,
-		PropertySetter,
-		EventAdd,
-		EventRemove
+		public readonly Interpreter Interpreter;
+
+		private readonly CSharp.ExpressionParser parser;
+
+		internal ExpressionParser (Interpreter interpreter)
+		{
+			this.Interpreter = interpreter;
+
+			parser = new CSharp.ExpressionParser ("C#");
+		}
+
+		public Expression Parse (string expression)
+		{
+			return parser.Parse (expression);
+		}
+
+		protected static SourceLocation FindFile (Thread target, string filename, int line)
+		{
+			SourceFile file = target.Process.FindFile (filename);
+			if (file == null)
+				throw new ScriptingException ("Cannot find source file `{0}'.",
+							      filename);
+
+			SourceMethod source = file.FindMethod (line);
+			if (source == null)
+				throw new ScriptingException (
+					"Cannot find method corresponding to line {0} in `{1}'.",
+					line, file.Name);
+
+			return new SourceLocation (source, line);
+		}
+
+		protected SourceLocation DoParseExpression (Thread target, LocationType type, string arg)
+		{
+			ScriptingContext context = new ScriptingContext (Interpreter);
+			context.CurrentFrame = target.CurrentFrame;
+
+			Expression expr = Interpreter.ExpressionParser.Parse (arg);
+			if (expr == null)
+				throw new ScriptingException ("Cannot resolve expression `{0}'.", arg);
+
+			MethodExpression mexpr;
+			try {
+				mexpr = expr.ResolveMethod (context, type);
+			} catch {
+				return null;
+			}
+
+			if (mexpr != null)
+				return mexpr.EvaluateSource (context);
+			else
+				return context.FindMethod (arg);
+		}
+
+		public static bool ParseLocation (Thread target, string arg, out SourceLocation location)
+		{
+			int line;
+			int pos = arg.IndexOf (':');
+			if (pos >= 0) {
+				string filename = arg.Substring (0, pos);
+				try {
+					line = (int) UInt32.Parse (arg.Substring (pos+1));
+				} catch {
+					throw new ScriptingException ("Expected filename:line");
+				}
+
+				location = FindFile (target, filename, line);
+				return true;
+			}
+
+			try {
+				line = (int) UInt32.Parse (arg);
+			} catch {
+				location = null;
+				return false;
+			}
+
+			StackFrame frame = target.CurrentFrame;
+			if ((frame.SourceAddress == null) || (frame.SourceAddress.Location == null))
+				throw new ScriptingException (
+					"Current stack frame doesn't have source code");
+
+			location = frame.SourceAddress.Location;
+			if (location.FileName == null)
+				throw new ScriptingException (
+					"Current stack frame doesn't have source code");
+
+			location = FindFile (target, location.FileName, line);
+			return true;
+		}
+
+		protected SourceLocation DoParse (Thread target, LocationType type, string arg)
+		{
+			if (type != LocationType.Default)
+				return DoParseExpression (target, type, arg);
+
+			SourceLocation location;
+			if (ParseLocation (target, arg, out location))
+				return location;
+
+			return DoParseExpression (target, type, arg);
+		}
+
+		public SourceLocation Parse (Thread target, LocationType type, string arg)
+		{
+			try {
+				return DoParse (target, type, arg);
+			} catch (ScriptingException ex) {
+				throw new TargetException (TargetError.LocationInvalid, ex.Message);
+			}
+
+			return null;
+		}
+
+		public string EvaluateExpression (ScriptingContext context, string text,
+						  DisplayFormat format)
+		{
+			Expression expression;
+
+			try {
+				expression = Interpreter.ExpressionParser.Parse (text);
+			} catch (ScriptingException ex) {
+				throw new ScriptingException ("Cannot parse expression `{0}': {1}.",
+							      text, ex.Message);
+			} catch {
+				throw new ScriptingException ("Cannot parse expression `{0}'.", text);
+			}
+
+			try {
+				expression = expression.Resolve (context);
+			} catch (ScriptingException ex) {
+				throw new ScriptingException ("Cannot resolve expression `{0}': {1}",
+							      text, ex.Message);
+			} catch {
+				throw new ScriptingException ("Cannot resolve expression `{0}'.", text);
+			}
+
+			try {
+				object retval = expression.Evaluate (context);
+				return context.FormatObject (retval, format);
+			} catch (ScriptingException ex) {
+				throw new ScriptingException ("Cannot evaluate expression `{0}': {1}",
+							      text, ex.Message);
+			} catch {
+				throw new ScriptingException ("Cannot evaluate expression `{0}'.", text);
+			}
+		}
 	}
 
 	public abstract class Expression
@@ -109,7 +250,23 @@ namespace Mono.Debugger.Frontend
 			if (var == null)
 				return null;
 
-			return var.GetObject (context.CurrentFrame);
+			TargetObject obj = var.GetObject (context.CurrentFrame);
+			if (obj == null)
+				return null;
+
+			TargetClassObject cobj = obj as TargetClassObject;
+			if (cobj != null) {
+				TargetObject current;
+				try {
+					current = cobj.GetCurrentObject (context.CurrentThread);
+				} catch {
+					current = null;
+				}
+				if (current != null)
+					return current;
+			}
+
+			return obj;
 		}
 
 		public TargetObject EvaluateObject (ScriptingContext context)
@@ -135,52 +292,15 @@ namespace Mono.Debugger.Frontend
 			}
 		}
 
-		protected virtual SourceLocation DoEvaluateSource (ScriptingContext context,
-								   LocationType type, Expression[] types)
-		{
-			TargetFunctionType func = DoEvaluateMethod (context, type, types);
-			if (func == null)
-				return null;
-
-			SourceMethod source = func.Source;
-			if (source == null)
-				return null;
-
-			return new SourceLocation (source);
-		}
-
-		public SourceLocation EvaluateSource (ScriptingContext context, LocationType type,
-						      Expression [] types)
-		{
-			if (!resolved)
-				throw new InvalidOperationException (
-					String.Format (
-						"Some clown tried to evaluate the " +
-						"unresolved expression `{0}'", Name));
-
-			try {
-				SourceLocation location = DoEvaluateSource (context, type, types);
-				if (location == null)
-					throw new ScriptingException (
-						"Expression `{0}' is not a method", Name);
-
-				return location;
-			} catch (LocationInvalidException ex) {
-				throw new ScriptingException (
-					"Location of variable `{0}' is invalid: {1}",
-					Name, ex.Message);
-			}
-		}
-
 		protected virtual TargetFunctionType DoEvaluateMethod (ScriptingContext context,
-									LocationType type,
-									Expression[] types)
+								       LocationType type,
+								       Expression[] types)
 		{
 			return null;
 		}
 
 		public TargetFunctionType EvaluateMethod (ScriptingContext context,
-							   LocationType type, Expression [] types)
+							  LocationType type, Expression [] types)
 		{
 			if (!resolved)
 				throw new InvalidOperationException (
@@ -259,14 +379,15 @@ namespace Mono.Debugger.Frontend
 			return expr;
 		}
 
-		protected virtual Expression DoResolveMethod (ScriptingContext context)
+		protected virtual MethodExpression DoResolveMethod (ScriptingContext context,
+								    LocationType type)
 		{
-			return DoResolve (context);
+			return null;
 		}
 
-		public Expression ResolveMethod (ScriptingContext context)
+		public MethodExpression ResolveMethod (ScriptingContext context, LocationType type)
 		{
-			Expression expr = DoResolveMethod (context);
+			MethodExpression expr = DoResolveMethod (context, type);
 			if (expr == null)
 				throw new ScriptingException (
 					"Expression `{0}' is not a method.", Name);
@@ -579,7 +700,15 @@ namespace Mono.Debugger.Frontend
 
 		protected override TargetObject DoEvaluateObject (ScriptingContext context)
 		{
-			TargetClassObject cobj = (TargetClassObject) base.DoEvaluateObject (context);
+			TargetVariable var = DoEvaluateVariable (context);
+			if (var == null)
+				return null;
+
+			TargetObject obj = var.GetObject (context.CurrentFrame);
+			if (obj == null)
+				return null;
+
+			TargetClassObject cobj = (TargetClassObject) obj;
 			return cobj.GetParentObject (context.CurrentThread);
 		}
 	}
@@ -716,15 +845,6 @@ namespace Mono.Debugger.Frontend
 			resolved = true;
 			return this;
 		}
-
-		protected override SourceLocation DoEvaluateSource (ScriptingContext context,
-								    LocationType type, Expression[] types)
-		{
-			if (types != null)
-				return null;
-
-			return location;
-		}
 	}
 
 	public class SimpleNameExpression : Expression
@@ -763,25 +883,24 @@ namespace Mono.Debugger.Frontend
 			if (member == null)
 				return null;
 
-			if (member.IsInstance && !method.HasThis)
-				throw new ScriptingException (
-					"Cannot use instance member `{0}' or current class " +
-					"in static context.", full_name);
-
 			return member;
 		}
 
-		Expression Lookup (ScriptingContext context, StackFrame frame)
+		MemberExpression Lookup (ScriptingContext context, StackFrame frame)
 		{
+			MemberExpression member = LookupMember (context, frame, name);
+			if (member != null)
+				return member;
+
 			string[] namespaces = context.GetNamespaces (frame);
 			if (namespaces == null)
 				return null;
 
 			foreach (string ns in namespaces) {
 				string full_name = MakeFQN (ns, name);
-				Expression expr = LookupMember (context, frame, full_name);
-				if (expr != null)
-					return expr;
+				member = LookupMember (context, frame, full_name);
+				if (member != null)
+					return member;
 			}
 
 			return null;
@@ -791,16 +910,13 @@ namespace Mono.Debugger.Frontend
 		{
 			StackFrame frame = context.CurrentFrame;
 			if (frame.Method != null) {
-				TargetVariable var = frame.Method.GetVariableByName (name);
+				TargetVariable var = frame.Method.GetVariableByName (
+					frame.TargetAddress, name);
 				if (var != null)
 					return new VariableAccessExpression (var);
 			}
 
-			Expression expr = LookupMember (context, frame, name);
-			if (expr != null)
-				return expr;
-
-			expr = Lookup (context, frame);
+			Expression expr = Lookup (context, frame);
 			if (expr != null)
 				return expr;
 
@@ -815,6 +931,33 @@ namespace Mono.Debugger.Frontend
 			throw new ScriptingException ("No symbol `{0}' in current context.", Name);
 		}
 
+		protected override MethodExpression DoResolveMethod (ScriptingContext context,
+								     LocationType type)
+		{
+			StackFrame frame = context.CurrentFrame;
+
+			MemberExpression member;
+			if (type == LocationType.Constructor) {
+				Expression texpr = ResolveType (context);
+				if (texpr == null)
+					return null;
+
+				TargetClassType ctype = texpr.EvaluateType (context) as TargetClassType;
+				if (ctype == null)
+					return null;
+
+				member = StructAccessExpression.FindMember (
+					frame.Thread, ctype, null, ".ctor", false, true);
+			} else {
+				member = Lookup (context, frame);
+			}
+
+			if (member != null)
+				return member.ResolveMethod (context, type);
+
+			return DoResolve (context) as MethodExpression;
+		}
+
 		protected override Expression DoResolveType (ScriptingContext context)
 		{
 			Language language = context.CurrentLanguage;
@@ -996,9 +1139,29 @@ namespace Mono.Debugger.Frontend
 			return ResolveMemberAccess (context, false);
 		}
 
-		protected override Expression DoResolveMethod (ScriptingContext context)
+		protected override MethodExpression DoResolveMethod (ScriptingContext context,
+								     LocationType type)
 		{
-			return ResolveMemberAccess (context, true);
+			MemberExpression member;
+			if (type == LocationType.Constructor) {
+				Expression texpr = ResolveType (context);
+				if (texpr == null)
+					return null;
+
+				TargetClassType ctype = texpr.EvaluateType (context) as TargetClassType;
+				if (ctype == null)
+					return null;
+
+				member = StructAccessExpression.FindMember (
+					context.CurrentFrame.Thread, ctype, null, ".ctor", false, true);
+			} else {
+				member = ResolveMemberAccess (context, true);
+			}
+
+			if (member != null)
+				return member.ResolveMethod (context, type);
+
+			return null;
 		}
 
 		protected override Expression DoResolveType (ScriptingContext context)
@@ -1037,7 +1200,34 @@ namespace Mono.Debugger.Frontend
 		}
 	}
 
-	public class MethodGroupExpression : MemberExpression
+	public abstract class MethodExpression : MemberExpression
+	{
+		protected abstract SourceLocation DoEvaluateSource (ScriptingContext context);
+
+		public SourceLocation EvaluateSource (ScriptingContext context)
+		{
+			if (!resolved)
+				throw new InvalidOperationException (
+					String.Format (
+						"Some clown tried to evaluate the " +
+						"unresolved expression `{0}'", Name));
+
+			try {
+				SourceLocation location = DoEvaluateSource (context);
+				if (location == null)
+					throw new ScriptingException (
+						"Expression `{0}' is not a method", Name);
+
+				return location;
+			} catch (LocationInvalidException ex) {
+				throw new ScriptingException (
+					"Location of variable `{0}' is invalid: {1}",
+					Name, ex.Message);
+			}
+		}
+	}
+
+	public class MethodGroupExpression : MethodExpression
 	{
 		protected readonly TargetClassType stype;
 		protected readonly TargetClassObject instance;
@@ -1079,6 +1269,12 @@ namespace Mono.Debugger.Frontend
 			return this;
 		}
 
+		protected override MethodExpression DoResolveMethod (ScriptingContext context,
+								     LocationType type)
+		{
+			return this;
+		}
+
 		protected override TargetObject DoEvaluateObject (ScriptingContext context)
 		{
 			throw new ScriptingException ("Expression `{0}' is a method, not a " +
@@ -1108,17 +1304,16 @@ namespace Mono.Debugger.Frontend
 			if (func != null)
 				return func;
 
-			ArrayList list = new ArrayList ();
-			foreach (TargetFunctionType method in methods) {
-				if (method.Source == null)
-					continue;
-				list.Add (method.Source);
-			}
-			if (list.Count == 0)
-				return null;
-			SourceMethod[] sources = new SourceMethod [list.Count];
-			list.CopyTo (sources, 0);
-			throw new MultipleLocationsMatchException (sources);
+			return context.Interpreter.QueryMethod (methods);
+		}
+
+		protected override SourceLocation DoEvaluateSource (ScriptingContext context)
+		{
+			if (methods.Length == 1)
+				return new SourceLocation ((TargetFunctionType) methods [0]);
+
+			throw new ScriptingException (
+				"Ambiguous method `{0}'; need to use full name", Name);
 		}
 
 		public TargetFunctionType OverloadResolve (ScriptingContext context,
@@ -1204,7 +1399,6 @@ namespace Mono.Debugger.Frontend
 		}
 	}
 
-#if FIXME
 	// So you can extend this by just creating a subclass
 	// of BinaryOperator that implements DoEvaluate and
 	// a constructor, but you'll need to add a new rule to
@@ -1217,7 +1411,7 @@ namespace Mono.Debugger.Frontend
 	// like supporting "a" + "b" = "ab", then larger changes would
 	// be needed.
 
-	public class BinaryOperator : Expression
+	public class BinaryOperator : PointerExpression
 	{
 		public enum Kind { Mult, Plus, Minus, Div };
 
@@ -1231,34 +1425,96 @@ namespace Mono.Debugger.Frontend
 			this.right = right;
 		}
 
-		protected object DoEvaluate (ScriptingContext context, object lobj, object robj)
+		public override string Name {
+			get {
+				string op;
+				switch (kind) {
+				case Kind.Mult:
+					op = "*";
+					break;
+				case Kind.Plus:
+					op = "+";
+					break;
+				case Kind.Minus:
+					op = "-";
+					break;
+				case Kind.Div:
+					op = "/";
+					break;
+				default:
+					throw new InternalError ();
+				}
+				return left.Name + op + right.Name;
+			}
+		}
+
+		protected long DoEvaluate (ScriptingContext context, long lvalue, long rvalue)
 		{
 			switch (kind) {
 			case Kind.Mult:
-				return (int) lobj * (int) robj;
+				return lvalue * rvalue;
 			case Kind.Plus:
-				return (int) lobj + (int) robj;
+				return lvalue + rvalue;
 			case Kind.Minus:
-				return (int) lobj - (int) robj;
+				return lvalue - rvalue;
 			case Kind.Div:
-				return (int) lobj / (int) robj;
+				return lvalue / rvalue;
 			}
 
 			throw new ScriptingException ("Unknown binary operator kind: {0}", kind);
 		}
 
+		private long GetValue (ScriptingContext context, Expression expr)
+		{
+			object val = expr.Evaluate (context);
+			if (val is int)
+				return (long) (int) val;
+			else if (val is uint)
+				return (long) (uint) val;
+			else if (val is ulong)
+				return (long) (ulong) val;
+			else if (val is long)
+				return (long) val;
+			else if (val is TargetPointerObject) {
+				TargetPointerObject pobj = (TargetPointerObject) val;
+				return pobj.Address.Address;
+			} else
+				throw new ScriptingException ("Cannot evaluate expression `{0}'", expr);
+		}
+
+		protected override object DoEvaluate (ScriptingContext context)
+		{
+			long lvalue = GetValue (context, left);
+			long rvalue = GetValue (context, right);
+
+			try {
+				long retval = DoEvaluate (context, lvalue, rvalue);
+				return new NumberExpression (retval);
+			} catch {
+				throw new ScriptingException ("Cannot evaluate expression `{0}'", Name);
+			}
+		}
+
+		public override TargetAddress EvaluateAddress (ScriptingContext context)
+		{
+			NumberExpression result = (NumberExpression) Evaluate (context);
+			return result.EvaluateAddress (context);
+		}
+
 		protected override Expression DoResolve (ScriptingContext context)
 		{
-			object lobj, robj;
+			left = left.Resolve (context);
+			if (left == null)
+				return null;
 
-			lobj = left.Resolve (context);
-			robj = right.Resolve (context);
+			right = right.Resolve (context);
+			if (right == null)
+				return null;
 
-			// Console.WriteLine ("bin eval: {0} ({1}) and {2} ({3})", lobj, lobj.GetType(), robj, robj.GetType());
-			return DoEvaluate (context, lobj, robj);
+			resolved = true;
+			return this;
 		}
 	}
-#endif
 
 	public class TypeOfExpression : Expression
 	{
@@ -1395,28 +1651,83 @@ namespace Mono.Debugger.Frontend
 			return this;
 		}
 
+		protected MethodGroupExpression CreateMethodGroup (TargetFunctionType func)
+		{
+			return new MethodGroupExpression (
+				func.DeclaringType, instance, func.Name,
+				new TargetFunctionType[] { func },
+				!func.IsStatic, func.IsStatic);
+		}
+
+		protected override MethodExpression DoResolveMethod (ScriptingContext context,
+								     LocationType type)
+		{
+			switch (type) {
+			case LocationType.PropertyGetter:
+			case LocationType.PropertySetter:
+				TargetPropertyInfo property = Member as TargetPropertyInfo;
+				if (property == null)
+					return null;
+
+				if (type == LocationType.PropertyGetter) {
+					if (!property.CanRead)
+						throw new ScriptingException (
+							"Property {0} doesn't have a getter.", Name);
+					return CreateMethodGroup (property.Getter);
+				} else {
+					if (!property.CanWrite)
+						throw new ScriptingException (
+							"Property {0} doesn't have a setter.", Name);
+					return CreateMethodGroup (property.Setter);
+				}
+
+			case LocationType.EventAdd:
+			case LocationType.EventRemove:
+				TargetEventInfo ev = Member as TargetEventInfo;
+				if (ev == null)
+					return null;
+
+				if (type == LocationType.EventAdd)
+					return CreateMethodGroup (ev.Add);
+				else
+					return CreateMethodGroup (ev.Remove);
+
+			case LocationType.Method:
+			case LocationType.DelegateInvoke:
+				TargetMethodInfo method = Member as TargetMethodInfo;
+				if (method != null)
+					return CreateMethodGroup (method.Type);
+
+				return InvocationExpression.ResolveDelegate (context, this);
+
+			default:
+				return null;
+			}
+		}
+
 		protected TargetObject GetField (Thread target, TargetFieldInfo field)
 		{
-			if (field.IsStatic)
+			if (field.HasConstValue)
+				return field.Type.Language.CreateInstance (target, field.ConstValue);
+			else if (field.IsStatic)
 				return Type.GetStaticField (target, field);
 			else
 				return InstanceObject.GetField (target, field);
 		}
 
 		protected TargetObject GetProperty (ScriptingContext context,
-						     TargetPropertyInfo prop)
+						    TargetPropertyInfo prop)
 		{
-			string exc_message;
-			TargetObject res = context.CurrentThread.RuntimeInvoke (
-				prop.Getter, InstanceObject, new TargetObject [0], true,
-				out exc_message);
+			RuntimeInvokeResult result = context.Interpreter.RuntimeInvoke (
+				context.CurrentThread, prop.Getter, InstanceObject,
+				new TargetObject [0], true, false);
 
-			if (exc_message != null)
+			if (result.ExceptionMessage != null)
 				throw new ScriptingException (
 					"Invocation of `{0}' raised an exception: {1}",
-					Name, exc_message);
+					Name, result.ExceptionMessage);
 
-			return res;
+			return result.ReturnObject;
 		}
 
 		protected TargetObject GetMember (ScriptingContext context, Thread target,
@@ -1510,6 +1821,11 @@ namespace Mono.Debugger.Frontend
 			}
 		}
 
+		protected override TargetType DoEvaluateType (ScriptingContext context)
+		{
+			return Member.Type;
+		}
+
 		public TargetFunctionType ResolveDelegate (ScriptingContext context)
 		{
 			MethodGroupExpression mg = InvocationExpression.ResolveDelegate (
@@ -1521,8 +1837,8 @@ namespace Mono.Debugger.Frontend
 		}
 
 		protected override TargetFunctionType DoEvaluateMethod (ScriptingContext context,
-									 LocationType type,
-									 Expression[] types)
+									LocationType type,
+									Expression[] types)
 		{
 			switch (type) {
 			case LocationType.PropertyGetter:
@@ -1652,19 +1968,30 @@ namespace Mono.Debugger.Frontend
 					result = null;
 				}
 
-				if (result == null)
-					throw new ScriptingException (
-						"Cannot dereference `{0}'.", expr.Name);
-
-				return result;
+				if (result != null)
+					return result;
 			}
 
 			TargetClassObject cobj = obj as TargetClassObject;
 			if (current_ok && (cobj != null))
 				return cobj;
 
+			PointerExpression pexpr = expr as PointerExpression;
+			if (pexpr != null) {
+				TargetAddress address = pexpr.EvaluateAddress (context);
+				TargetAddress result;
+				try {
+					result = context.CurrentThread.ReadAddress (address);
+					return context.CurrentFrame.Language.CreatePointer (
+						context.CurrentFrame, result);
+				} catch {
+					throw new ScriptingException ("Cannot dereference `{0}'.",
+								      expr.Name);
+				}
+			}
+
 			throw new ScriptingException (
-				"Expression `{0}' is not a pointer type.", expr.Name);
+				"Expression `{0}' is not a pointer.", expr.Name);
 		}
 
 		public override TargetAddress EvaluateAddress (ScriptingContext context)
@@ -1674,6 +2001,8 @@ namespace Mono.Debugger.Frontend
 				obj = (long) (int) obj;
 			if (obj is long)
 				return new TargetAddress (context.AddressDomain, (long) obj);
+			else if (obj is PointerExpression)
+				return ((PointerExpression) obj).EvaluateAddress (context);
 
 			TargetPointerObject pobj = obj as TargetPointerObject;
 			if (pobj == null)
@@ -1825,7 +2154,7 @@ namespace Mono.Debugger.Frontend
 				int[] int_indices = GetIntIndices (target, context);
 				try {
 					return aobj.GetElement (target, int_indices);
-				} catch (ArgumentException ex) {
+				} catch (ArgumentException) {
 					throw new ScriptingException (
 						"Index of array expression `{0}' out of bounds.",
 						expr.Name);
@@ -1904,7 +2233,7 @@ namespace Mono.Debugger.Frontend
 				int[] int_indices = GetIntIndices (target, context);
 				try {
 					aobj.SetElement (target, int_indices, right);
-				} catch (ArgumentException ex) {
+				} catch (ArgumentException) {
 					throw new ScriptingException (
 						"Index of array expression `{0}' out of bounds.",
 						expr.Name);
@@ -2035,6 +2364,32 @@ namespace Mono.Debugger.Frontend
 			return TryCurrentCast (context, target, sobj, target_type);
 		}
 
+		static bool TryParentCast (ScriptingContext context,
+					   TargetClassType source_type,
+					   TargetClassType target_type)
+		{
+			if (source_type == target_type)
+				return true;
+
+			if (!source_type.HasParent)
+				return false;
+
+			return TryParentCast (context, source_type.ParentType, target_type);
+		}
+
+		public static bool TryCast (ScriptingContext context, TargetType source,
+					    TargetClassType target_type)
+		{
+			if (source == target_type)
+				return true;
+
+			TargetClassType stype = Convert.ToClassType (source);
+			if (stype == null)
+				return false;
+
+			return TryParentCast (context, stype, target_type);
+		}
+
 		protected override TargetObject DoEvaluateObject (ScriptingContext context)
 		{
 			TargetType target_type = target.EvaluateType (context);
@@ -2328,6 +2683,12 @@ namespace Mono.Debugger.Frontend
 					return ctype;
 			}
 
+			TargetArrayType atype = type as TargetArrayType;
+			if (atype != null) {
+				if (atype.Language.ArrayType != null)
+					return atype.Language.ArrayType;
+			}
+
 			throw new ScriptingException (
 				"Type `{0}' is not a struct or class.", type.Name);
 		}
@@ -2342,6 +2703,10 @@ namespace Mono.Debugger.Frontend
 			if (oobj != null)
 				return oobj.GetClassObject (target);
 
+			TargetArrayObject aobj = obj as TargetArrayObject;
+			if ((aobj != null) && aobj.HasClassObject)
+				return aobj.GetClassObject (target);
+
 			return null;
 		}
 	}
@@ -2414,22 +2779,22 @@ namespace Mono.Debugger.Frontend
 		}
 	}
 
-	public class InvocationExpression : Expression
+	public class InvocationExpression : MethodExpression
 	{
-		Expression method_expr;
+		Expression expr;
 		Expression[] arguments;
-		MethodGroupExpression mg;
+		MethodGroupExpression method_expr;
 		string name;
 
 		TargetType[] argtypes;
 		TargetFunctionType method;
 
-		public InvocationExpression (Expression method_expr, Expression[] arguments)
+		public InvocationExpression (Expression expr, Expression[] arguments)
 		{
-			this.method_expr = method_expr;
+			this.expr = expr;
 			this.arguments = arguments;
 
-			name = String.Format ("{0} ()", method_expr.Name);
+			name = String.Format ("{0} ()", expr.Name);
 		}
 
 		public override string Name {
@@ -2439,17 +2804,24 @@ namespace Mono.Debugger.Frontend
 		public static MethodGroupExpression ResolveDelegate (ScriptingContext context,
 								     Expression expr)
 		{
-			TargetClassObject sobj = Convert.ToClassObject (
-				context.CurrentThread, expr.EvaluateObject (context));
-			if (sobj == null)
+			TargetClassType ctype = Convert.ToClassType (expr.EvaluateType (context));
+			if (ctype == null)
 				return null;
 
-			TargetClassType delegate_type = sobj.Type.Language.DelegateType;
-			if (CastExpression.TryCast (context, sobj, delegate_type) == null)
+			TargetClassObject cobj;
+			try {
+				cobj = Convert.ToClassObject (
+					context.CurrentThread, expr.EvaluateObject (context));
+			} catch {
+				cobj = null;
+			}
+
+			TargetClassType delegate_type = ctype.Language.DelegateType;
+			if (!CastExpression.TryCast (context, ctype, delegate_type))
 				return null;
 
 			TargetFunctionType invoke = null;
-			foreach (TargetMethodInfo method in sobj.Type.Methods) {
+			foreach (TargetMethodInfo method in ctype.Methods) {
 				if (method.Name == "Invoke") {
 					invoke = method.Type;
 					break;
@@ -2462,24 +2834,40 @@ namespace Mono.Debugger.Frontend
 			TargetFunctionType[] methods = new TargetFunctionType[] { invoke };
 
 			MethodGroupExpression mg = new MethodGroupExpression (
-				sobj.Type, sobj, "Invoke", methods, true, false);
+				ctype, cobj, "Invoke", methods, true, false);
 			return mg;
 		}
 
+		public override TargetClassObject InstanceObject {
+			get { return method_expr.InstanceObject; }
+		}
+
+		public override bool IsInstance {
+			get { return method_expr.IsInstance; }
+		}
+
+		public override bool IsStatic {
+			get { return method_expr.IsStatic; }
+		}
+
+		protected override MethodExpression DoResolveMethod (ScriptingContext context,
+								     LocationType type)
+		{
+			method_expr = (MethodGroupExpression) expr.ResolveMethod (context, type);
+			if (method_expr == null)
+				return null;
+
+			resolved = true;
+			return this;
+		}
+
 		protected override Expression DoResolve (ScriptingContext context)
 		{
-			method_expr = method_expr.Resolve (context);
+			method_expr = (MethodGroupExpression) expr.ResolveMethod (
+				context, LocationType.Method);
 			if (method_expr == null)
 				return null;
 
-			mg = method_expr as MethodGroupExpression;
-			if (mg == null)
-				mg = ResolveDelegate (context, method_expr);
-
-			if (mg == null)
-				throw new ScriptingException (
-					"Expression `{0}' is not a method.", Name);
-
 			argtypes = new TargetType [arguments.Length];
 
 			for (int i = 0; i < arguments.Length; i++) {
@@ -2490,7 +2878,7 @@ namespace Mono.Debugger.Frontend
 				argtypes [i] = arguments [i].EvaluateType (context);
 			}
 
-			method = mg.OverloadResolve (context, argtypes);
+			method = method_expr.OverloadResolve (context, argtypes);
 
 			resolved = true;
 			return this;
@@ -2513,12 +2901,29 @@ namespace Mono.Debugger.Frontend
 		}
 
 		protected override TargetFunctionType DoEvaluateMethod (ScriptingContext context,
-									 LocationType type,
-									 Expression[] types)
+									LocationType type,
+									Expression[] types)
 		{
 			return method_expr.EvaluateMethod (context, type, types);
 		}
 
+		protected override SourceLocation DoEvaluateSource (ScriptingContext context)
+		{
+			Expression[] types = new Expression [arguments.Length];
+			TargetType[] argtypes = new TargetType [types.Length];
+
+			for (int i = 0; i < arguments.Length; i++) {
+				types [i] = arguments [i].ResolveType (context);
+				argtypes [i] = types [i].EvaluateType (context);
+			}
+
+			TargetFunctionType func = method_expr.OverloadResolve (context, argtypes);
+			if (func != null)
+				return new SourceLocation (func);
+
+			return null;
+		}
+
 		protected TargetObject DoInvoke (ScriptingContext context, bool debug)
 		{
 			TargetObject[] args = new TargetObject [arguments.Length];
@@ -2532,29 +2937,30 @@ namespace Mono.Debugger.Frontend
 					context, args [i], method.ParameterTypes [i]);
 			}
 
-			TargetClassObject instance = mg.InstanceObject;
+			TargetClassObject instance = method_expr.InstanceObject;
+
+			if (!method.IsStatic && !method.IsConstructor && (instance == null))
+				throw new ScriptingException (
+					"Cannot invoke instance method `{0}' with a type reference.",
+					method.FullName);
 
 			try {
-				if (debug) {
-					Thread thread = context.CurrentThread;
-					thread.RuntimeInvoke (method, instance, objs, true);
+				Thread thread = context.CurrentThread;
+				RuntimeInvokeResult result;
 
-					if (context.Interpreter.IsSynchronous)
-						context.Interpreter.Wait (thread);
+				result = context.Interpreter.RuntimeInvoke (
+					thread, method, instance, objs, true, debug);
 
-					return null;
-				}
+				if (result == null)
+					throw new ScriptingException (
+						"Invocation of `{0}' aborted abnormally.", Name);
 
-				string exc_message;
-				TargetObject retval = context.CurrentThread.RuntimeInvoke (
-					method, mg.InstanceObject, objs, true, out exc_message);
-
-				if (exc_message != null)
+				if (result.ExceptionMessage != null)
 					throw new ScriptingException (
 						"Invocation of `{0}' raised an exception: {1}",
-						Name, exc_message);
+						Name, result.ExceptionMessage);
 
-				return retval;
+				return result.ReturnObject;
 			} catch (TargetException ex) {
 				throw new ScriptingException (
 					"Invocation of `{0}' raised an exception: {1}", Name, ex.Message);
