@@ -1,14 +1,23 @@
 using System;
+using Mono.Debugger.Backends;
 
 namespace Mono.Debugger.Languages
 {
-	public abstract class Language : MarshalByRefObject
+	public abstract class Language : DebuggerMarshalByRefObject
 	{
 		public abstract string Name {
 			get;
 		}
 
-		public abstract ITargetInfo TargetInfo {
+		public abstract bool IsManaged {
+			get;
+		}
+
+		internal abstract ProcessServant Process {
+			get;
+		}
+
+		public abstract TargetInfo TargetInfo {
 			get;
 		}
 
@@ -44,6 +53,10 @@ namespace Mono.Debugger.Languages
 			get;
 		}
 
+		public abstract TargetClassType ArrayType {
+			get;
+		}
+
 		public abstract string SourceLanguage (StackFrame frame);
 
 		public abstract TargetType LookupType (string name);
