@@ -41,183 +41,165 @@ namespace Mono.Debugger.Languages.Mono
 		public readonly MonoStringType StringType;
 		public readonly MonoClassType ExceptionType;
 		public readonly MonoClassType DelegateType;
+		public readonly MonoClassType ArrayType;
 
-		public readonly int TypeSize;
-		public readonly int ArrayTypeSize;
-		public readonly int KlassSize;
-		public readonly int KlassInstanceSizeOffset;
-		public readonly int KlassParentOffset;
-		public readonly int KlassTokenOffset;
-		public readonly int KlassFieldOffset;
-		public readonly int KlassMethodsOffset;
-		public readonly int KlassMethodCountOffset;
-		public readonly int KlassThisArgOffset;
-		public readonly int KlassByValArgOffset;
-		public readonly int KlassGenericClassOffset;
-		public readonly int KlassGenericContainerOffset;
-		public readonly int FieldInfoSize;
-
-		public MonoBuiltinTypeInfo (MonoSymbolFile corlib, ITargetMemoryAccess memory,
-					    TargetAddress address)
+		public MonoBuiltinTypeInfo (MonoSymbolFile corlib, TargetMemoryAccess memory,
+					    MonoMetadataInfo info)
 		{
 			this.Corlib = corlib;
 
-			int size = memory.ReadInteger (address);
-			TargetBinaryReader reader = memory.ReadMemory (address, size).GetReader ();
-			reader.ReadInt32 ();
-
-			int defaults_size = reader.ReadInt32 ();
-			TargetAddress defaults_address = new TargetAddress (
-				memory.AddressDomain, reader.ReadAddress ());
-
-			TypeSize = reader.ReadInt32 ();
-			ArrayTypeSize = reader.ReadInt32 ();
-			KlassSize = reader.ReadInt32 ();
-			KlassInstanceSizeOffset = reader.ReadInt32 ();
-			KlassParentOffset = reader.ReadInt32 ();
-			KlassTokenOffset = reader.ReadInt32 ();
-			KlassFieldOffset = reader.ReadInt32 ();
-			KlassMethodsOffset = reader.ReadInt32 ();
-			KlassMethodCountOffset = reader.ReadInt32 ();
-			KlassThisArgOffset = reader.ReadInt32 ();
-			KlassByValArgOffset = reader.ReadInt32 ();
-			KlassGenericClassOffset = reader.ReadInt32 ();
-			KlassGenericContainerOffset = reader.ReadInt32 ();
-			FieldInfoSize = reader.ReadInt32 ();
-
 			TargetReader mono_defaults = new TargetReader (
-				memory.ReadMemory (defaults_address, defaults_size).Contents, memory);
-			mono_defaults.ReadAddress ();
+				memory.ReadMemory (info.MonoDefaultsAddress, info.MonoDefaultsSize));
 
 			MonoLanguageBackend language = corlib.MonoLanguage;
+			mono_defaults.Offset = info.MonoDefaultsObjectOffset;
 			TargetAddress klass = mono_defaults.ReadAddress ();
 			int object_size = 2 * corlib.TargetInfo.TargetAddressSize;
-			Cecil.ITypeDefinition object_type = corlib.ModuleDefinition.Types ["System.Object"];
+			Cecil.TypeDefinition object_type = corlib.ModuleDefinition.Types ["System.Object"];
 			ObjectType = new MonoObjectType (corlib, object_type, object_size);
-			ObjectClass = new MonoClassType (corlib, object_type);
+			ObjectClass = new MonoClassType (memory, corlib, object_type, klass);
 			language.AddCoreType (ObjectClass, object_type, klass);
 
+			mono_defaults.Offset = info.MonoDefaultsByteOffset;
 			klass = mono_defaults.ReadAddress ();
-			Cecil.ITypeDefinition byte_type = corlib.ModuleDefinition.Types ["System.Byte"];
+			Cecil.TypeDefinition byte_type = corlib.ModuleDefinition.Types ["System.Byte"];
 			ByteType = new TargetFundamentalType (language, byte_type.FullName, FundamentalKind.Byte, 1);
 			language.AddCoreType (ByteType, byte_type, klass);
 
+			mono_defaults.Offset = info.MonoDefaultsVoidOffset;
 			klass = mono_defaults.ReadAddress ();
-			Cecil.ITypeDefinition void_type = corlib.ModuleDefinition.Types ["System.Void"];
+			Cecil.TypeDefinition void_type = corlib.ModuleDefinition.Types ["System.Void"];
 			VoidType = new MonoOpaqueType (corlib, void_type);
 			language.AddCoreType (VoidType, void_type, klass);
 
+			mono_defaults.Offset = info.MonoDefaultsBooleanOffset;
 			klass = mono_defaults.ReadAddress ();
-			Cecil.ITypeDefinition bool_type = corlib.ModuleDefinition.Types ["System.Boolean"];
+			Cecil.TypeDefinition bool_type = corlib.ModuleDefinition.Types ["System.Boolean"];
 			BooleanType = new TargetFundamentalType (language, bool_type.FullName, FundamentalKind.Boolean, 1);
 			language.AddCoreType (BooleanType, bool_type, klass);
 
+			mono_defaults.Offset = info.MonoDefaultsSByteOffset;
 			klass = mono_defaults.ReadAddress ();
-			Cecil.ITypeDefinition sbyte_type = corlib.ModuleDefinition.Types ["System.SByte"];
+			Cecil.TypeDefinition sbyte_type = corlib.ModuleDefinition.Types ["System.SByte"];
 			SByteType = new TargetFundamentalType (language, sbyte_type.FullName, FundamentalKind.SByte, 1);
 			language.AddCoreType (SByteType, sbyte_type, klass);
 
+			mono_defaults.Offset = info.MonoDefaultsInt16Offset;
 			klass = mono_defaults.ReadAddress ();
-			Cecil.ITypeDefinition short_type = corlib.ModuleDefinition.Types ["System.Int16"];
+			Cecil.TypeDefinition short_type = corlib.ModuleDefinition.Types ["System.Int16"];
 			Int16Type = new TargetFundamentalType (language, short_type.FullName, FundamentalKind.Int16, 2);
 			language.AddCoreType (Int16Type, short_type, klass);
 
+			mono_defaults.Offset = info.MonoDefaultsUInt16Offset;
 			klass = mono_defaults.ReadAddress ();
-			Cecil.ITypeDefinition ushort_type = corlib.ModuleDefinition.Types ["System.UInt16"];
+			Cecil.TypeDefinition ushort_type = corlib.ModuleDefinition.Types ["System.UInt16"];
 			UInt16Type = new TargetFundamentalType (language, ushort_type.FullName, FundamentalKind.UInt16, 2);
 			language.AddCoreType (UInt16Type, ushort_type, klass);
 
+			mono_defaults.Offset = info.MonoDefaultsInt32Offset;
 			klass = mono_defaults.ReadAddress ();
-			Cecil.ITypeDefinition int_type = corlib.ModuleDefinition.Types ["System.Int32"];
+			Cecil.TypeDefinition int_type = corlib.ModuleDefinition.Types ["System.Int32"];
 			Int32Type = new TargetFundamentalType (language, int_type.FullName, FundamentalKind.Int32, 4);
 			language.AddCoreType (Int32Type, int_type, klass);
 
+			mono_defaults.Offset = info.MonoDefaultsUInt32Offset;
 			klass = mono_defaults.ReadAddress ();
-			Cecil.ITypeDefinition uint_type = corlib.ModuleDefinition.Types ["System.UInt32"];
+			Cecil.TypeDefinition uint_type = corlib.ModuleDefinition.Types ["System.UInt32"];
 			UInt32Type = new TargetFundamentalType (language, uint_type.FullName, FundamentalKind.UInt32, 4);
 			language.AddCoreType (UInt32Type, uint_type, klass);
 
+			mono_defaults.Offset = info.MonoDefaultsIntOffset;
 			klass = mono_defaults.ReadAddress ();
-			Cecil.ITypeDefinition intptr_type = corlib.ModuleDefinition.Types ["System.IntPtr"];
+			Cecil.TypeDefinition intptr_type = corlib.ModuleDefinition.Types ["System.IntPtr"];
 			IntType = new TargetFundamentalType (language, intptr_type.FullName, FundamentalKind.IntPtr, 4);
 			language.AddCoreType (IntType, intptr_type, klass);
 
+			mono_defaults.Offset = info.MonoDefaultsUIntOffset;
 			klass = mono_defaults.ReadAddress ();
-			Cecil.ITypeDefinition uintptr_type = corlib.ModuleDefinition.Types ["System.UIntPtr"];
+			Cecil.TypeDefinition uintptr_type = corlib.ModuleDefinition.Types ["System.UIntPtr"];
 			UIntType = new TargetFundamentalType (language, uintptr_type.FullName, FundamentalKind.Object, 4);
 			language.AddCoreType (UIntType, uintptr_type, klass);
 
+			mono_defaults.Offset = info.MonoDefaultsInt64Offset;
 			klass = mono_defaults.ReadAddress ();
-			Cecil.ITypeDefinition long_type = corlib.ModuleDefinition.Types ["System.Int64"];
+			Cecil.TypeDefinition long_type = corlib.ModuleDefinition.Types ["System.Int64"];
 			Int64Type = new TargetFundamentalType (language, long_type.FullName, FundamentalKind.Int64, 8);
 			language.AddCoreType (Int64Type, long_type, klass);
 
+			mono_defaults.Offset = info.MonoDefaultsUInt64Offset;
 			klass = mono_defaults.ReadAddress ();
-			Cecil.ITypeDefinition ulong_type = corlib.ModuleDefinition.Types ["System.UInt64"];
+			Cecil.TypeDefinition ulong_type = corlib.ModuleDefinition.Types ["System.UInt64"];
 			UInt64Type = new TargetFundamentalType (language, ulong_type.FullName, FundamentalKind.UInt64, 8);
 			language.AddCoreType (UInt64Type, ulong_type, klass);
 
+			mono_defaults.Offset = info.MonoDefaultsSingleOffset;
 			klass = mono_defaults.ReadAddress ();
-			Cecil.ITypeDefinition float_type = corlib.ModuleDefinition.Types ["System.Single"];
+			Cecil.TypeDefinition float_type = corlib.ModuleDefinition.Types ["System.Single"];
 			SingleType = new TargetFundamentalType (language, float_type.FullName, FundamentalKind.Single, 4);
 			language.AddCoreType (SingleType, float_type, klass);
 
+			mono_defaults.Offset = info.MonoDefaultsDoubleOffset;
 			klass = mono_defaults.ReadAddress ();
-			Cecil.ITypeDefinition double_type = corlib.ModuleDefinition.Types ["System.Double"];
+			Cecil.TypeDefinition double_type = corlib.ModuleDefinition.Types ["System.Double"];
 			DoubleType = new TargetFundamentalType (language, double_type.FullName, FundamentalKind.Double, 8);
 			language.AddCoreType (DoubleType, double_type, klass);
 
+			mono_defaults.Offset = info.MonoDefaultsCharOffset;
 			klass = mono_defaults.ReadAddress ();
-			Cecil.ITypeDefinition char_type = corlib.ModuleDefinition.Types ["System.Char"];
+			Cecil.TypeDefinition char_type = corlib.ModuleDefinition.Types ["System.Char"];
 			CharType = new TargetFundamentalType (language, char_type.FullName, FundamentalKind.Char, 2);
 			language.AddCoreType (CharType, char_type, klass);
 
+			mono_defaults.Offset = info.MonoDefaultsStringOffset;
 			klass = mono_defaults.ReadAddress ();
-			Cecil.ITypeDefinition string_type = corlib.ModuleDefinition.Types ["System.String"];
+			Cecil.TypeDefinition string_type = corlib.ModuleDefinition.Types ["System.String"];
 			StringType = new MonoStringType (
 				corlib, string_type.FullName, object_size, object_size + 4);
 			language.AddCoreType (StringType, string_type, klass);
 
-			// Skip a whole bunch of clases we don't care about
-			mono_defaults.Offset += 2 * corlib.TargetInfo.TargetAddressSize;
-
+			mono_defaults.Offset = info.MonoDefaultsArrayOffset;
 			klass = mono_defaults.ReadAddress ();
-			Cecil.ITypeDefinition delegate_type = corlib.ModuleDefinition.Types ["System.Delegate"];
+			Cecil.TypeDefinition array_type = corlib.ModuleDefinition.Types ["System.Array"];
+			ArrayType = new MonoClassType (memory, corlib, array_type, klass);
+			language.AddCoreType (ArrayType, array_type, klass);
+
+			mono_defaults.Offset = info.MonoDefaultsDelegateOffset;
+			klass = mono_defaults.ReadAddress ();
+			Cecil.TypeDefinition delegate_type = corlib.ModuleDefinition.Types ["System.Delegate"];
 			DelegateType = new MonoClassType (corlib, delegate_type);
 			language.AddCoreType (DelegateType, delegate_type, klass);
 
-			// Skip a whole bunch of clases we don't care about
-			mono_defaults.Offset += 7 * corlib.TargetInfo.TargetAddressSize;
-
-			// and get to the Exception class
+			mono_defaults.Offset = info.MonoDefaultsExceptionOffset;
 			klass = mono_defaults.ReadAddress ();
-			Cecil.ITypeDefinition exception_type = corlib.ModuleDefinition.Types ["System.Exception"];
-			ExceptionType = new MonoClassType (corlib, exception_type);
+			Cecil.TypeDefinition exception_type = corlib.ModuleDefinition.Types ["System.Exception"];
+			ExceptionType = new MonoClassType (memory, corlib, exception_type, klass);
 			language.AddCoreType (ExceptionType, exception_type, klass);
 		}
 	}
 
 	internal class MonoLanguageBackend : Language, ILanguageBackend
 	{
-		ArrayList symbol_files;
+		Hashtable symbol_files;
 		int last_num_symbol_files;
 		Hashtable image_hash;
 		Hashtable assembly_hash;
 		Hashtable assembly_by_name;
 		Hashtable class_hash;
+		Hashtable class_info_by_addr;
+		Hashtable class_info_by_type;
 		MonoSymbolFile corlib;
 		MonoBuiltinTypeInfo builtin_types;
 
 		int last_num_data_tables;
 		int last_data_table_offset;
 
-		Process process;
+		ProcessServant process;
 		MonoDebuggerInfo info;
 		TargetAddress[] trampolines;
 		bool initialized;
 		DebuggerMutex mutex;
 
-		public MonoLanguageBackend (Process process, MonoDebuggerInfo info)
+		public MonoLanguageBackend (ProcessServant process, MonoDebuggerInfo info)
 		{
 			this.process = process;
 			this.info = info;
@@ -228,10 +210,18 @@ namespace Mono.Debugger.Languages.Mono
 			get { return "Mono"; }
 		}
 
+		public override bool IsManaged {
+			get { return true; }
+		}
+
 		internal MonoDebuggerInfo MonoDebuggerInfo {
 			get { return info; }
 		}
 
+		internal MonoMetadataInfo MonoMetadataInfo {
+			get { return info.MonoMetadataInfo; }
+		}
+
 		Language ILanguageBackend.Language {
 			get { return this; }
 		}
@@ -240,17 +230,21 @@ namespace Mono.Debugger.Languages.Mono
 			get { return builtin_types; }
 		}
 
+		internal override ProcessServant Process {
+			get { return process; }
+		}
+
 		public SourceFileFactory SourceFileFactory {
 			get { return process.SourceFileFactory; }
 		}
 
-		public override ITargetInfo TargetInfo {
+		public override TargetInfo TargetInfo {
 			get { return corlib.TargetInfo; }
 		}
 
 		internal bool TryFindImage (Thread thread, string filename)
 		{
-			Cecil.IAssemblyDefinition ass = Cecil.AssemblyFactory.GetAssembly (filename);
+			Cecil.AssemblyDefinition ass = Cecil.AssemblyFactory.GetAssembly (filename);
 			if (ass == null)
 				return false;
 
@@ -261,11 +255,11 @@ namespace Mono.Debugger.Languages.Mono
 			return true;
 		}
 
-		public TargetType LookupMonoType (Cecil.ITypeReference type)
+		public TargetType LookupMonoType (Cecil.TypeReference type)
 		{
 			MonoSymbolFile file;
 
-			Cecil.ITypeDefinition typedef = type as Cecil.ITypeDefinition;
+			Cecil.TypeDefinition typedef = type as Cecil.TypeDefinition;
 			if (typedef != null) {
 				file = (MonoSymbolFile) assembly_hash [type.Module.Assembly];
 				if (file == null) {
@@ -277,7 +271,7 @@ namespace Mono.Debugger.Languages.Mono
 				return file.LookupMonoType (typedef);
 			}
 
-			Cecil.IArrayType array = type as Cecil.IArrayType;
+			Cecil.ArrayType array = type as Cecil.ArrayType;
 			if (array != null) {
 				TargetType element_type = LookupMonoType (array.ElementType);
 				if (element_type == null)
@@ -286,7 +280,7 @@ namespace Mono.Debugger.Languages.Mono
 				return new MonoArrayType (element_type, array.Rank);
 			}
 
-			Cecil.IReferenceType reftype = type as Cecil.IReferenceType;
+			Cecil.ReferenceType reftype = type as Cecil.ReferenceType;
 			if (reftype != null) {
 				TargetType element_type = LookupMonoType (reftype.ElementType);
 				if (element_type == null)
@@ -322,26 +316,31 @@ namespace Mono.Debugger.Languages.Mono
 				return mono_type;
 		}
 
-		public void AddClass (TargetAddress klass_address, TargetType type)
-		{
-			if (!class_hash.Contains (klass_address))
-				class_hash.Add (klass_address, type);
-		}
-
-		internal void AddCoreType (TargetType type, Cecil.ITypeDefinition typedef,
+		internal void AddCoreType (TargetType type, Cecil.TypeDefinition typedef,
 					   TargetAddress klass)
 		{
 			corlib.AddType (type, typedef);
-			AddClass (klass, type);
+			if (!class_hash.Contains (klass))
+				class_hash.Add (klass, type);
 		}
 
-		public TargetType GetClass (Thread target, TargetAddress klass_address)
+		public TargetType GetClass (TargetMemoryAccess target, TargetAddress klass_address)
 		{
 			TargetType type = (TargetType) class_hash [klass_address];
 			if (type != null)
 				return type;
 
-			return MonoClassType.ReadMonoClass (this, target, klass_address);
+			try {
+				type = MonoClassType.ReadMonoClass (this, target, klass_address);
+			} catch {
+				return null;
+			}
+
+			if (type == null)
+				return null;
+
+			class_hash.Add (klass_address, type);
+			return type;
 		}
 
 		public MonoSymbolFile GetImage (TargetAddress address)
@@ -349,60 +348,62 @@ namespace Mono.Debugger.Languages.Mono
 			return (MonoSymbolFile) image_hash [address];
 		}
 
-		void read_mono_debugger_info (ITargetMemoryAccess memory)
-		{
-			init_trampolines (memory);
-
-			symbol_files = new ArrayList ();
-			image_hash = new Hashtable ();
-			assembly_hash = new Hashtable ();
-			assembly_by_name = new Hashtable ();
-			class_hash = new Hashtable ();
-		}
-
-		void init_trampolines (ITargetMemoryAccess memory)
+		void read_mono_debugger_info (TargetMemoryAccess memory)
 		{
 			trampolines = new TargetAddress [4];
 			TargetAddress address = info.MonoTrampolineCode;
 			trampolines [0] = memory.ReadAddress (address);
-			address += memory.TargetAddressSize;
+			address += memory.TargetInfo.TargetAddressSize;
 			trampolines [1] = memory.ReadAddress (address);
-			address += memory.TargetAddressSize;
+			address += memory.TargetInfo.TargetAddressSize;
 			trampolines [2] = memory.ReadAddress (address);
-			address += 2 * memory.TargetAddressSize;
+			address += 2 * memory.TargetInfo.TargetAddressSize;
 			trampolines [3] = memory.ReadAddress (address);
+
+			symbol_files = new Hashtable ();
+			image_hash = new Hashtable ();
+			assembly_hash = new Hashtable ();
+			assembly_by_name = new Hashtable ();
+			class_hash = new Hashtable ();
+			class_info_by_addr = new Hashtable ();
+			class_info_by_type = new Hashtable ();
+			initialized = true;
 		}
 
 #region symbol table management
-		internal void Update (ITargetMemoryAccess target)
+		internal void Update (TargetMemoryAccess target)
 		{
 			do_update_symbol_table (target);
 		}
 
-		void do_update_symbol_table (ITargetMemoryAccess memory)
+		void do_update_symbol_table (TargetMemoryAccess memory)
 		{
+			if (!initialized)
+				return;
+
 			Report.Debug (DebugFlags.JitSymtab, "Starting to update symbol table");
-			process.ModuleManager.Lock ();
 			try {
 				do_update (memory);
 			} catch (ThreadAbortException) {
 				return;
 			} catch (Exception e) {
-				Console.WriteLine ("Can't update symbol table: {0}", e);
+				Console.WriteLine ("Can't update symbol table: {0} {1} {2}",
+						   memory, e, Environment.StackTrace);
 				return;
-			} finally {
-				process.ModuleManager.UnLock ();
 			}
 			Report.Debug (DebugFlags.JitSymtab, "Done updating symbol table");
 		}
 
 		// This method reads the MonoDebuggerSymbolTable structure
 		// (struct definition is in mono-debug-debugger.h)
-		void do_update (ITargetMemoryAccess memory)
+		void do_update (TargetMemoryAccess memory)
 		{
 			TargetAddress symtab_address = memory.ReadAddress (info.SymbolTable);
+			if (symtab_address.IsNull)
+				return;
+
 			TargetReader header = new TargetReader (
-				memory.ReadMemory (symtab_address, info.SymbolTableSize), memory);
+				memory.ReadMemory (symtab_address, info.SymbolTableSize));
 
 			long magic = header.BinaryReader.ReadInt64 ();
 			if (magic != MonoDebuggerInfo.DynamicMagic)
@@ -432,30 +433,41 @@ namespace Mono.Debugger.Languages.Mono
 			TargetAddress symfiles_address = header.ReadAddress ();
 			int num_symbol_files = header.ReadInteger ();
 
-			symfiles_address += last_num_symbol_files * memory.TargetAddressSize;
+			symfiles_address += last_num_symbol_files * memory.TargetInfo.TargetAddressSize;
 			for (int i = last_num_symbol_files; i < num_symbol_files; i++) {
 				TargetAddress address = memory.ReadAddress (symfiles_address);
-				symfiles_address += memory.TargetAddressSize;
+				symfiles_address += memory.TargetInfo.TargetAddressSize;
+
+				MonoSymbolFile symfile = null;
 
 				try {
-					MonoSymbolFile symfile = new MonoSymbolFile (
-						this, process, memory, memory, address);
-					image_hash.Add (symfile.MonoImage, symfile);
-					symbol_files.Add (symfile);
-					assembly_hash.Add (symfile.Assembly, symfile);
-					assembly_by_name.Add (symfile.Assembly.Name.FullName, symfile);
-
-					if (address != corlib_address)
-						continue;
-
-					corlib = symfile;
-					builtin_types = new MonoBuiltinTypeInfo (
-						corlib, memory, info.MetadataInfo);
+					symfile = new MonoSymbolFile (this, process, memory, address);
 				} catch (C.MonoSymbolFileException ex) {
 					Console.WriteLine (ex.Message);
+				} catch (SymbolTableException ex) {
+					Console.WriteLine (ex.Message);
 				} catch (Exception ex) {
 					Console.WriteLine (ex);
 				}
+
+				if (symfile == null)
+					continue;
+
+				if (assembly_by_name.Contains (symfile.Assembly.Name.FullName))
+					continue;
+
+				image_hash.Add (symfile.MonoImage, symfile);
+				assembly_hash.Add (symfile.Assembly, symfile);
+				assembly_by_name.Add (symfile.Assembly.Name.FullName, symfile);
+
+				symbol_files.Add (i, symfile);
+
+				if (address != corlib_address)
+					continue;
+
+				corlib = symfile;
+				builtin_types = new MonoBuiltinTypeInfo (
+					corlib, memory, info.MonoMetadataInfo);
 			}
 
 			last_num_symbol_files = num_symbol_files;
@@ -463,7 +475,7 @@ namespace Mono.Debugger.Languages.Mono
 		}
 
 		// This method reads a portion of the data table (defn in mono-debug.h)
-		void read_data_table (ITargetMemoryAccess memory, ITargetMemoryReader header)
+		void read_data_table (TargetMemoryAccess memory, TargetReader header)
 		{
 			int num_data_tables = header.ReadInteger ();
 			TargetAddress data_tables = header.ReadAddress ();
@@ -474,11 +486,11 @@ namespace Mono.Debugger.Languages.Mono
 			if (num_data_tables != last_num_data_tables) {
 				int old_offset = last_data_table_offset;
 
-				data_tables += last_num_data_tables * memory.TargetAddressSize;
+				data_tables += last_num_data_tables * memory.TargetInfo.TargetAddressSize;
 
 				for (int i = last_num_data_tables; i < num_data_tables; i++) {
 					TargetAddress old_table = memory.ReadAddress (data_tables);
-					data_tables += memory.TargetAddressSize;
+					data_tables += memory.TargetInfo.TargetAddressSize;
 
 					int old_size = memory.ReadInteger (old_table);
 					read_data_items (memory, old_table, old_offset, old_size);
@@ -513,17 +525,17 @@ namespace Mono.Debugger.Languages.Mono
 			Wrapper
 		}
 
-		void read_data_items (ITargetMemoryAccess memory, TargetAddress address, int start, int end)
+		void read_data_items (TargetMemoryAccess memory, TargetAddress address, int start, int end)
 		{
 			TargetReader reader = new TargetReader (
-				memory.ReadMemory (address + start, end - start), memory);
+				memory.ReadMemory (address + start, end - start));
 
 			Report.Debug (DebugFlags.JitSymtab,
 				      "READ DATA ITEMS: {0} {1} {2} - {3} {4}", address, start, end,
 				      reader.BinaryReader.Position, reader.Size);
 
 			if (start == 0)
-				reader.BinaryReader.Position = memory.TargetAddressSize;
+				reader.BinaryReader.Position = memory.TargetInfo.TargetAddressSize;
 
 			while (reader.BinaryReader.Position + 4 < reader.Size) {
 				int item_size = reader.BinaryReader.ReadInt32 ();
@@ -539,7 +551,7 @@ namespace Mono.Debugger.Languages.Mono
 					break;
 
 				case DataItemType.Class:
-					read_class_entry (reader);
+					read_class_entry (memory, reader);
 					break;
 
 				case DataItemType.Wrapper:
@@ -551,39 +563,77 @@ namespace Mono.Debugger.Languages.Mono
 			}
 		}
 
-		void read_range_entry (ITargetMemoryReader reader)
-		{
-			int size = reader.BinaryReader.PeekInt32 ();
-			byte[] contents = reader.BinaryReader.PeekBuffer (size);
-			reader.BinaryReader.ReadInt32 ();
-			MonoSymbolFile file = (MonoSymbolFile) symbol_files [reader.BinaryReader.ReadInt32 ()];
-			file.AddRangeEntry (reader, contents);
-		}
-
-		void read_class_entry (ITargetMemoryReader reader)
+		void read_range_entry (TargetReader reader)
 		{
 			int size = reader.BinaryReader.PeekInt32 ();
 			byte[] contents = reader.BinaryReader.PeekBuffer (size);
 			reader.BinaryReader.ReadInt32 ();
 			int file_idx = reader.BinaryReader.ReadInt32 ();
+			Report.Debug (DebugFlags.JitSymtab, "READ RANGE ITEM: {0} {1}",
+				      size, file_idx);
+			MonoSymbolFile file = (MonoSymbolFile) symbol_files [file_idx];
+			if (file != null)
+				file.AddRangeEntry (reader, contents);
+		}
 
-			if (file_idx >= symbol_files.Count)
-				return;
+		void read_class_entry (TargetMemoryAccess memory, TargetReader reader)
+		{
+			reader.BinaryReader.ReadInt32 ();
+			int file_idx = reader.BinaryReader.ReadInt32 ();
 
 			MonoSymbolFile file = (MonoSymbolFile) symbol_files [file_idx];
 			if (file == null)
 				return;
 
-			file.AddClassEntry (reader, contents);
+			reader.BinaryReader.ReadLeb128 ();
+			reader.BinaryReader.ReadLeb128 ();
+			TargetAddress klass_address = reader.ReadAddress ();
+
+			GetClassInfo (memory, klass_address);
 		}
 
-		void read_wrapper_entry (ITargetMemoryAccess memory, ITargetMemoryReader reader)
+		void read_wrapper_entry (TargetMemoryAccess memory, TargetReader reader)
 		{
 			int size = reader.BinaryReader.PeekInt32 ();
 			byte[] contents = reader.BinaryReader.PeekBuffer (size);
 			reader.BinaryReader.ReadInt32 ();
 			corlib.AddWrapperEntry (memory, reader, contents);
 		}
+
+		internal MonoClassInfo GetClassInfo (TargetMemoryAccess memory, TargetAddress klass_address)
+		{
+			MonoClassInfo info = (MonoClassInfo) class_info_by_addr [klass_address];
+			if (info != null)
+				return info;
+
+			int token = memory.ReadInteger (
+				klass_address + MonoMetadataInfo.KlassTokenOffset);
+
+			TargetAddress image = memory.ReadAddress (klass_address);
+			MonoSymbolFile file = GetImage (image);
+			if (file == null)
+				return null;
+
+			if ((token & 0xff000000) != 0x02000000)
+				throw new InternalError ();
+			token &= 0x00ffffff;
+
+			Cecil.TypeDefinition tdef = file.ModuleDefinition.LookupByToken (
+				Cecil.Metadata.TokenType.TypeDef, (int) token) as Cecil.TypeDefinition;
+			if (tdef == null)
+				throw new InternalError ();
+
+			info = new MonoClassInfo (memory, this, token, klass_address);
+			class_info_by_addr.Add (klass_address, info);
+			if (!info.IsGenericClass)
+				class_info_by_type.Add (tdef, info);
+			return info;
+		}
+
+		internal MonoClassInfo GetClassInfo (Cecil.TypeDefinition type)
+		{
+			return (MonoClassInfo) class_info_by_type [type];
+		}
 #endregion
 
 #region jit breakpoint handling
@@ -648,14 +698,17 @@ namespace Mono.Debugger.Languages.Mono
 			if (name.IndexOf ('[') >= 0)
 				return null;
 
-			foreach (MonoSymbolFile symfile in symbol_files) {
-				Cecil.ITypeDefinitionCollection types = symfile.Assembly.MainModule.Types;
-				// FIXME: Work around an API problem in Cecil.
-				foreach (Cecil.ITypeDefinition type in types) {
-					if (type.FullName != name)
-						continue;
+			foreach (MonoSymbolFile symfile in symbol_files.Values) {
+				try {
+					Cecil.TypeDefinitionCollection types = symfile.Assembly.MainModule.Types;
+					// FIXME: Work around an API problem in Cecil.
+					foreach (Cecil.TypeDefinition type in types) {
+						if (type.FullName != name)
+							continue;
 
-					return symfile.LookupMonoType (type);
+						return symfile.LookupMonoType (type);
+					}
+				} catch {
 				}
 			}
 
@@ -784,6 +837,10 @@ namespace Mono.Debugger.Languages.Mono
 		public override TargetClassType ObjectType {
 			get { return builtin_types.ObjectClass; }
 		}
+
+		public override TargetClassType ArrayType {
+			get { return builtin_types.ArrayType; }
+		}
 #endregion
 
 #region ILanguageBackend implementation
@@ -791,7 +848,7 @@ namespace Mono.Debugger.Languages.Mono
 			get { return info.RuntimeInvoke; }
 		}
 
-		public TargetAddress GetTrampolineAddress (ITargetMemoryAccess memory,
+		public TargetAddress GetTrampolineAddress (TargetMemoryAccess memory,
 							   TargetAddress address,
 							   out bool is_start)
 		{
@@ -830,7 +887,7 @@ namespace Mono.Debugger.Languages.Mono
 			get { return info.RunFinally; }
 		}
 
-		public SourceMethod GetTrampoline (ITargetMemoryAccess memory,
+		public SourceMethod GetTrampoline (TargetMemoryAccess memory,
 						   TargetAddress address)
 		{
 			bool is_start;
@@ -842,7 +899,7 @@ namespace Mono.Debugger.Languages.Mono
 			TargetAddress klass = memory.ReadAddress (trampoline + 8);
 			TargetAddress image = memory.ReadAddress (klass);
 
-			foreach (MonoSymbolFile file in symbol_files) {
+			foreach (MonoSymbolFile file in symbol_files.Values) {
 				if (file.MonoImage != image)
 					continue;
 
@@ -886,8 +943,8 @@ namespace Mono.Debugger.Languages.Mono
 				break;
 
 			default:
-				Console.WriteLine ("Received unknown notification {0:x}",
-						   (int) type);
+				Console.WriteLine ("Received unknown notification {0:x} / {1} {2:x}",
+						   (int) type, data, arg);
 				break;
 			}
 		}
@@ -906,7 +963,7 @@ namespace Mono.Debugger.Languages.Mono
 
 			if (disposing) {
 				if (symbol_files != null) {
-					foreach (MonoSymbolFile symfile in symbol_files)
+					foreach (MonoSymbolFile symfile in symbol_files.Values)
 						symfile.Dispose();
 
 					symbol_files = null;
